

from option_chain_analytics.data.config import (NearestStrikeOnGrid,
                                                StrikeSelection,
                                                OptionTickerConfig,
                                                compute_time_to_maturity,
                                                compute_days_to_maturity,
                                                mat_to_timestamp,
                                                split_option_contract_ticker,
                                                get_option_data_from_contract,
                                                get_option_data_from_contracts,
                                                FutureTickerConfig,
                                                split_future_contract_ticker,
                                                get_ttm_from_future_ticker,
                                                get_file_name)


from option_chain_analytics.data.chain_loader_from_ts import (create_chain_from_from_options_dfs,
                                                                           create_chain_timeseries,
                                                                           generate_atm_vols,
                                                                           generate_vol_delta_ts)

from option_chain_analytics.data.chain_ts import (ChainTs,
                                                  FuturesChainTs,
                                                  OptionsDataDFs)




import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import qis
from enum import Enum

# analytics
from option_chain_analytics import OptionsDataDFs
from option_chain_analytics.ts_loaders import ts_data_loader_wrapper, DataSource


pd.set_option('display.max_rows', 500)
pd.set_option('display.max_columns', 500)
pd.set_option('display.width', 1000)


class UnitTests(Enum):
    PLOT_FUNDING_RATE = 1


def run_unit_test(unit_test: UnitTests):

    ticker = 'BTC'  # BTC, ETH

    options_data_dfs = OptionsDataDFs(**ts_data_loader_wrapper(ticker=ticker, data_source=DataSource.TARDIS_LOCAL))

    if unit_test == UnitTests.PLOT_FUNDING_RATE:

        daily_1h = qis.generate_dates_schedule(time_period=options_data_dfs.get_start_end_date(), freq='H')

        daily_8h = qis.generate_dates_schedule(time_period=options_data_dfs.get_start_end_date(), freq='D',
                                               hour_offset=8)

        # deribit extrapolates funding rate by 8.0
        funding_rate_1h = options_data_dfs.get_funding_rate(freq=None, is_rescale_to_one_hour=True)
        funding_rate_1 = options_data_dfs.get_funding_rate(freq='H', index=daily_1h)
        funding_rate_2 = options_data_dfs.get_funding_rate(freq='D', index=daily_8h)
        print(f"funding_rate_1=\n{funding_rate_1}")
        print(f"funding_rate_2=\n{funding_rate_2}")

        cumrates = pd.concat([funding_rate_1h.cumsum(0).rename('1h'),
                              funding_rate_1.cumsum(0).rename('H at index'),
                              funding_rate_2.cumsum(0).rename('D at index')],
                             axis=1).dropna()

        funding_rate_annual = 365.0*funding_rate_1h.resample('D').sum()

        with sns.axes_style("darkgrid"):
            fig, axs = plt.subplots(3, 1, figsize=(10, 7))
            kwargs = dict(x_date_freq='M', legend_stats=qis.LegendStats.FIRST_AVG_LAST, framealpha=0.9)
            qis.plot_time_series(df=funding_rate_1h, title='funding_rate', ax=axs[0], **kwargs)
            qis.plot_time_series(df=funding_rate_annual, title='funding_rate_annual', ax=axs[1], **kwargs)
            qis.plot_time_series(df=cumrates, title='cum_funding_rate', ax=axs[2], **kwargs)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.PLOT_FUNDING_RATE

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from typing import Tuple, Optional, List, Dict
import qis


def plot_vols(vols: pd.DataFrame,
              ax: plt.Subplot = None,
              linestyles: List[str] = None,
              markers: List[str] = None,
              label_x_y: Dict[str, Tuple[float, float]] = None,
              **kwargs
              ) -> None:

    kwargs = qis.update_kwargs(kwargs,
                               dict(ncol=3, legend_loc='upper center',
                                    xvar_format='{:,.2f}',
                                    yvar_format='{:.0%}',
                                    markersize=10))
    if ax is None:
        with sns.axes_style("darkgrid"):
            fig, ax = plt.subplots(1, 1, figsize=(6, 6), tight_layout=True)

    markers = markers or ["o"]*len(vols.columns)
    linestyles = linestyles or ['-']*len(vols.columns)
    qis.plot_line(df=vols,
                  linestyles=linestyles,
                  markers=markers,
                  ylabel='Implied Vol',
                  ax=ax,
                  **kwargs)

    if label_x_y is not None:
        qis.add_scatter_points(ax=ax, label_x_y=label_x_y, linewidth=10)


def map_deltas_to_str(bsm_deltas: np.ndarray, delta_str_format: str = '0.2f') -> List[str]:
    """
    map deltas to str of 0.2f
    deltas below 0.05 are mapped as 0.04
    """
    slice_index = []
    index_str = np.empty_like(bsm_deltas, dtype=str)
    for idx, x in enumerate(bsm_deltas):
        if np.abs(x) < 0.05:
            x_str = f"{x:0.4f}"
        else:
            x_str = f"{x:{delta_str_format}}"
        # check for non overlaps
        if idx > 0:
            if x_str == index_str[idx - 1]:
                if x < 0.0:  # decrease previous delta
                    slice_index[idx-1] = f"{bsm_deltas[idx-1]:0.4f}"
                else:
                    x_str = f"{x:0.4f}"
        slice_index.append(x_str)
    return slice_index



import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import qis
from typing import Dict
from enum import Enum

# analytics
from option_chain_analytics.data.chain_loader_from_ts import create_chain_from_from_options_dfs
from option_chain_analytics.data.chain_ts import OptionsDataDFs
import option_chain_analytics.visuals.slices as vis
from option_chain_analytics.option_chain import SlicesChain

FIG_SIZE = (8.3, 11.7)  # A4 for portrait


def run_chain_report(chain: SlicesChain) -> Dict[str, plt.Figure]:
    figs = {}
    configs = {'Unrestricted': dict(delta_bounds=None, is_filtered=False),
               'Deltas > 0.1': dict(delta_bounds=(-0.1, 0.1), is_filtered=True)}
    for expiry, eslice in chain.expiry_slices.items():
        if eslice.get_ttm() > 0.0:
            for key, vals in configs.items():
                fig = plt.figure(figsize=FIG_SIZE, constrained_layout=True)
                fig.suptitle(f"slice id={eslice.expiry_id}, future price={eslice.forward:,.2f} - {key}",
                             fontweight="bold", fontsize=10, color='blue')
                gs = fig.add_gridspec(nrows=2, ncols=1, wspace=0.0, hspace=0.0)
                with sns.axes_style("darkgrid"):
                    ax = fig.add_subplot(gs[0, 0])
                    vis.plot_slice_vols_with_oi(eslice=eslice, title=f"{eslice.expiry_id} Vols In Strike Space",
                                                is_delta_space=False,
                                                ax=ax, **vals)
                    ax = fig.add_subplot(gs[1, 0])
                    vis.plot_slice_vols_with_oi(eslice=eslice, title=f"{eslice.expiry_id} Vols in Delta Space",
                                                is_delta_space=True,
                                                ax=ax, **vals)
                figs[f"{expiry}_{key}"] = fig

    plt.close('all')

    return figs


class UnitTests(Enum):
    RUN_CHAIN_REPORT = 1


def run_unit_test(unit_test: UnitTests):

    import local_path as local_path
    from option_chain_analytics.ts_loaders import ts_data_loader_wrapper, DataSource

    ticker = 'BTC'
    options_data_dfs = OptionsDataDFs(**ts_data_loader_wrapper(ticker=ticker, data_source=DataSource.TARDIS_LOCAL))

    if unit_test == UnitTests.RUN_CHAIN_REPORT:
        value_time = pd.Timestamp('2023-02-06 08:00:00+00:00')
        chain = create_chain_from_from_options_dfs(options_data_dfs=options_data_dfs, value_time=value_time)
        figs = run_chain_report(chain=chain)

        qis.save_figs_to_pdf(figs=figs,
                             file_name=f"chain_report_{value_time:%Y%m%dT%H%M%S}",
                             orientation='landscape',
                             local_path=local_path.get_output_path())

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.RUN_CHAIN_REPORT

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
define api and data specific conversions of options and futures tickers
"""

import re
import numpy as np
import pandas as pd
from typing import Tuple, Optional, List
from enum import Enum

TIME_FMT = '%Y%m%d%H%M%S'
EXPIRY_DATE_FORMAT = '%d%b%Y'
SECONDS_PER_DAY = 24*60*60  # hours, minute, seconds


class NearestStrikeOnGrid(Enum):
    # nearest strike from the strikes grid, for put ratios must be strictly below or above
    NEAREST = 0
    MAX_OI = 1
    ABOVE = 2
    BELOW = 3


class StrikeSelection(Enum):
    # strike selection for a strategy
    ATM = 1
    DELTA = 2


class OptionTickerConfig(Enum):
    """
    enumerate ticker methodology for each exchange or data provider
    we need to extract 3 data from the ticker: maturity, strike, type
    """
    CMS = 1   # deribit-ETH-28OCT22-2000-C-option
    DERIBIT = 2  # ETH-13JAN23-1100-P
    YAHOO = 3  # SPY230117C00353000


def compute_time_to_maturity(maturity_time: pd.Timestamp,
                             value_time: pd.Timestamp,
                             is_floor_at_zero: bool = True,
                             af: float = 365
                             ) -> float:
    """
    return annualised difference between mat_date and value_time
    """
    ttm = (maturity_time - value_time).total_seconds() / (af*SECONDS_PER_DAY)
    if is_floor_at_zero and ttm < 0.0:
        ttm = 0.0
    return ttm


def compute_days_to_maturity(maturity_time: pd.Timestamp,
                             value_time: pd.Timestamp,
                             is_floor_at_zero: bool = True,
                             af: float = 365
                             ) -> float:
    """
    return annualised difference between mat_date and value_time
    """
    ttm = (maturity_time - value_time).total_seconds() / (af*SECONDS_PER_DAY)
    if is_floor_at_zero and ttm < 0.0:
        ttm = 0.0
    return ttm


def mat_to_timestamp(mat: str,
                     date_format: Optional[str] = None,  # deribit = "%d%b%y"
                     hour_offset: Optional[int] = 8
                     ) -> pd.Timestamp:
    """
    from maturity string get timestamp
    """
    mat = pd.Timestamp(pd.to_datetime(mat, format=date_format), tz='UTC')
    if hour_offset is not None:
        mat = mat + pd.to_timedelta(hour_offset, unit='h')
    return mat


def split_option_contract_ticker(contract: str = 'deribit-ETH-28OCT22-2000-C-option',
                                 option_ticker_config: OptionTickerConfig = OptionTickerConfig.CMS
                                 ) -> Tuple[str, float, str]:
    """
    split ticker as follows
    deribit-SOL-28OCT22-10-C-option
    """
    if option_ticker_config == OptionTickerConfig.CMS:
        parts = contract.split('-')
        if len(parts) < 5:
            raise ValueError(f"invalid contract = {contract}")
        mat_str = parts[2]
        strike = float(parts[3])
        option_type = parts[4]
    elif option_ticker_config == OptionTickerConfig.DERIBIT:
        parts = contract.split('-')
        mat_str = parts[1]
        strike = float(parts[2])
        option_type = parts[3]
    else:
        raise NotImplementedError

    return mat_str, strike, option_type


def get_option_data_from_contract(value_time: pd.Timestamp,
                                  contract: str = 'ETH-28OCT22-2000-C-option',
                                  option_ticker_config: OptionTickerConfig = OptionTickerConfig.DERIBIT
                                  ) -> Tuple[str, pd.Timestamp, float, str, float]:
    """
    split ticker as follows
    deribit-SOL-28OCT22-10-C-option
    """
    mat_str, strike, option_type = split_option_contract_ticker(contract=contract,
                                                                option_ticker_config=option_ticker_config)
    mat_date = mat_to_timestamp(mat_str)
    ttm = compute_time_to_maturity(maturity_time=mat_date, value_time=value_time)
    return mat_str, mat_date, strike, option_type, ttm


def get_option_data_from_contracts(value_times: List[pd.Timestamp],
                                   contracts: List[str],
                                   option_ticker_config: OptionTickerConfig = OptionTickerConfig.DERIBIT
                                   ) -> pd.DataFrame:
    """
    split ticker as follows
    deribit-SOL-28OCT22-10-C-option
    """
    rows = {}
    for idx, (value_time, contract) in enumerate(zip(value_times, contracts)):
        mat_str, mat_date, strike, option_type, ttm = get_option_data_from_contract(value_time=value_time,
                                                                                    contract=contract,
                                                                                    option_ticker_config=option_ticker_config)
        rows[idx] = dict(mat_str=mat_str, mat_date=mat_date, strike=strike, option_type=option_type, ttm=ttm)
    df = pd.DataFrame.from_dict(rows, orient='index')
    return df


class FutureTickerConfig(Tuple, Enum):
    """
    enumertate ticker methodology for each exchange or data provider
    we need to extract 3 data from the ticker: maturity, strike, type
    """
    CMS = (1, "%d%b%y")  # deribit-ETH-28OCT22-future
    DERIBIT = (2, "%d%b%y")  # ETH-13JAN23
    BINANCEUSDM = (3, "%y%m%d")  # ETH/USDT, ETHUSDT_230331
    BYBIT = (4, "%y%m%d")  # ETH/USD:ETH-230630
    OKX = (5, "%y%m%d")  # ETH-USDT-SWAP
    HUOBI = (6, "%y%m%d")  # ETH-USD, ETH-USDT-230127


def split_future_contract_ticker(contract: str = 'deribit-ETH-30DEC22-future',
                                 future_ticker_config: FutureTickerConfig = FutureTickerConfig.CMS
                                 ) -> Tuple[str, Optional[pd.Timestamp]]:
    """
    split ticker as follows
    deribit-SOL-28OCT22-10-C-option
    """
    if future_ticker_config == FutureTickerConfig.CMS:
        parts = contract.split('-')
        mat_str = parts[2]
        if mat_str == 'PERPETUAL':
            mat_ts = None
        else:
            mat_ts = mat_to_timestamp(mat_str, date_format=future_ticker_config[1])

    elif future_ticker_config == FutureTickerConfig.DERIBIT:
        parts = contract.split('-')
        mat_str = parts[-1]
        if mat_str == 'PERPETUAL':
            mat_ts = None
        else:
            mat_ts = mat_to_timestamp(mat_str, date_format=future_ticker_config[1])

    elif future_ticker_config == FutureTickerConfig.BINANCEUSDM:
        parts = contract.split('_')
        mat_str = parts[-1]
        if len(parts) == 1:
            mat_ts = None
        else:
            mat_ts = mat_to_timestamp(mat_str, date_format=future_ticker_config[1])

    elif future_ticker_config == FutureTickerConfig.BYBIT:
        parts = contract.split('-')
        mat_str = parts[-1]
        if len(parts) == 1:  # perp is /USD:USDC
            mat_ts = None
        else:
            mat_ts = mat_to_timestamp(mat_str, date_format=future_ticker_config[1])

    elif future_ticker_config == FutureTickerConfig.OKX:
        parts = contract.split('-')
        mat_str = parts[-1]
        if mat_str == 'SWAP':
            mat_ts = None
        else:
            mat_ts = mat_to_timestamp(mat_str, date_format=future_ticker_config[1])

    elif future_ticker_config == FutureTickerConfig.HUOBI:
        parts = contract.split('-')
        mat_str = parts[-1]
        if not len(parts) == 3:
            mat_ts = None
        else:
            mat_str = re.findall(r'\d+', contract)[0]  # get list of nums in string
            mat_ts = mat_to_timestamp(mat_str, date_format=future_ticker_config[1])
    else:
        raise NotImplementedError(f"future_ticker_config={future_ticker_config}")

    return mat_str, mat_ts


def get_ttm_from_future_ticker(value_time: pd.Timestamp,
                               contract: str = 'deribit-ETH-30DEC22-future',
                               future_ticker_config: FutureTickerConfig = FutureTickerConfig.CMS,
                               is_floor_at_zero: bool = True
                               ) -> float:
    mat_str, mat_date = split_future_contract_ticker(contract=contract, future_ticker_config=future_ticker_config)
    if mat_date is not None:
        ttm = compute_time_to_maturity(maturity_time=mat_date, value_time=value_time, is_floor_at_zero=is_floor_at_zero)
    else:
        if is_floor_at_zero:  # so no confusion
            ttm = 0.0
        else:
            ttm = np.inf
    return ttm


def get_file_name(ticker: str, freq: Optional[str], hour_offset: Optional[int]) -> str:
    if freq is not None:
        if freq == 'H':
            file_name = f"{ticker}_freq_{freq}"
        else:
            if hour_offset is not None:
                file_name = f"{ticker}_freq_{freq}_hour_{hour_offset}"
            else:
                file_name = f"{ticker}_freq_{freq}"
    else:
        file_name = ticker
    return file_name


class UnitTests(Enum):
    OPTIONS_TICKER = 1
    FUTURES_TICKER = 2


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.OPTIONS_TICKER:
        mat_str, strike, option_type = split_option_contract_ticker()
        print(mat_str)
        print(strike)
        print(option_type)

    elif unit_test == UnitTests.FUTURES_TICKER:
        mat_str, mat_ts = split_future_contract_ticker(contract='deribit-BTC-18MAR22-future')
        print(f"mat_str={mat_str}")
        print(f"mat_ts={mat_ts}")

        mat_str, mat_ts = split_future_contract_ticker(contract='ETH/USD:USDC',
                                                       future_ticker_config=FutureTickerConfig.BYBIT)
        print(f"mat_str={mat_str}")
        print(f"mat_ts={mat_ts}")

        mat_str, mat_ts = split_future_contract_ticker(contract='ETHUSDT_230331',
                                                       future_ticker_config=FutureTickerConfig.BINANCEUSDM)
        print(f"mat_str={mat_str}")
        print(f"mat_ts={mat_ts}")


if __name__ == '__main__':

    unit_test = UnitTests.FUTURES_TICKER

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
old_analytics for fitting vol for log sv model
"""
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit, brenth
from scipy.stats import norm
from typing import Tuple, Optional, Dict, Union
from enum import Enum

import option_chain_analytics.pricers.bsm as bsm

FUTURES = 'futures'
ATM_VOL = 'sigma0'
BETA = 'beta'
VOLVOL = 'volvol'


def cals_logsv_atm_fit(log_strikes: np.ndarray,
                       mid_vols: np.ndarray,
                       strike_step: float = 0.3
                       ) -> Dict[str, float]:
    """
    compute atm fit for beta for initial guesses for optimisation
    """
    atm_vol = np.interp(x=0.0, xp=log_strikes, fp=mid_vols)
    atm_vol_m1 = np.interp(x=-strike_step, xp=log_strikes, fp=mid_vols)
    atm_vol_p1 = np.interp(x=strike_step, xp=log_strikes, fp=mid_vols)
    beta = (atm_vol_m1 - atm_vol_p1) / (2.0 * strike_step)
    convexity = np.maximum((atm_vol_m1 - 2.0 * atm_vol + atm_vol_p1) / (strike_step * strike_step), 0.01)
    volvol = np.sqrt(0.5 * (12.0 * (convexity * atm_vol) + beta * beta))
    atm_fit_params = {ATM_VOL: atm_vol, BETA: beta, VOLVOL: volvol}
    return atm_fit_params


def fit_logsv_ivols(log_strikes: np.ndarray,
                    mid_vols: np.ndarray,
                    ttm: float,
                    is_vega_weights: bool = True
                    ) -> Dict[str, float]:
    """
    fit logsv vol for zero mean-reversion
    """
    bounds = ([0.01, -15.0, 0.01], [np.nanmax(mid_vols), 5.0, 30.0])
    atm_fit_params = cals_logsv_atm_fit(log_strikes=log_strikes, mid_vols=mid_vols)
    # p0 = np.array([atm_fit_params[ATM_VOL], atm_fit_params[BETA], atm_fit_params[VOLVOL]])
    p0 = np.array([atm_fit_params[ATM_VOL], 0.0, 0.1])

    # f(x; params)
    def func(log_strikes, sigma0, beta, volvol):
        return calc_logsv_ivols(log_strikes, sigma0, beta, volvol)

    # sigma is the inverse of vega weights
    if is_vega_weights:
        vol = mid_vols * np.sqrt(ttm)
        d1 = -log_strikes / vol + 0.5 * vol
        vega = np.sqrt(ttm) * np.exp(-0.5 * d1 * d1)
        sigma = np.reciprocal(vega)
    else:
        sigma = None
    popt, pcov = curve_fit(f=func,
                           xdata=log_strikes,
                           ydata=mid_vols,
                           bounds=bounds,
                           p0=p0,
                           sigma=sigma)
    fit_params = {ATM_VOL: popt[0], BETA: popt[1], VOLVOL: popt[2]}
    return fit_params


def calc_logsv_ivols(log_strikes: Union[float, np.ndarray],
                     sigma0: float,
                     beta: float,
                     volvol: float,
                     is_quadratic: bool = True
                     ) -> Union[float, np.ndarray]:
    y = - log_strikes / sigma0
    # for approximation near zero
    b = - beta / 2.0
    c = 2.0 * volvol * volvol - beta * beta
    quadratic_ivols = sigma0 * (1.0 + (b + c * y) * y)
    if is_quadratic:
        return quadratic_ivols
    else:
        vartheta2 = beta * beta + volvol * volvol
        vartheta = np.sqrt(vartheta2)
        j_y = np.sqrt(1.0 + vartheta2 * y * y - 2.0 * beta * y)
        x = (1.0 / vartheta) * np.log((j_y * vartheta + vartheta2 * y - beta) / (vartheta - beta))
        ivols = np.divide(y, x, where=np.greater(np.abs(x), 0.0))
        ivols = np.where(np.greater(np.abs(x), 0.0), ivols,  quadratic_ivols)
    return ivols


def calc_logsv_ivols_partials(log_strikes: np.ndarray,
                              sigma0: float,
                              beta: float,
                              volvol: float,
                              eps: float = 0.01,
                              mult: float = 1.0,
                              is_analytic: bool = False
                              ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    if not is_analytic:
        def sigma(x: np.ndarray) -> np.ndarray:
            return calc_logsv_ivols(log_strikes=x, sigma0=sigma0, beta=beta, volvol=volvol)

        sigma_0 = sigma(x=log_strikes)
        sigma_p = sigma(x=log_strikes + eps)
        sigma_m = sigma(x=log_strikes - eps)
        dsigma = (0.5 / eps) * (sigma_p - sigma_m)
        d2sigma = (1.0 / (eps * eps)) * (sigma_p - 2.0 * sigma_0 + sigma_m)
        return mult * sigma_0, mult * dsigma, mult * d2sigma
    else:
        y = - log_strikes / sigma0
        vartheta2 = beta * beta + volvol * volvol
        vartheta = np.sqrt(vartheta2)
        j_y = np.sqrt(1.0 + vartheta2 * y * y - 2.0 * beta * y)
        log_y = np.log((j_y * vartheta + vartheta2 * y - beta) / (vartheta - beta))
        impvol = sigma0 * y / (1 / vartheta * log_y)
        DimpvolDy = -vartheta2 * sigma0 * y / (j_y * log_y * log_y) + sigma0 * vartheta / log_y
        D2impvolDy2 = vartheta2 * sigma0 * (2 * y * vartheta * j_y - (j_y * j_y - y * beta + 1) * log_y) / np.power(
            j_y * log_y, 3.0)
        DimpvolDk = DimpvolDy * (-1.0 / sigma0)
        D2impvolDk2 = D2impvolDy2 * (1.0 / sigma0 / sigma0)
        return mult * impvol, mult * DimpvolDk, mult * D2impvolDk2


def calc_logsv_pdf(ttm: float,
                   sigma0: float,
                   beta: float,
                   volvol: float,
                   log_strikes: Optional[np.ndarray] = None,
                   is_norm: bool = False,
                   cut: float = 6.0,
                   is_analytic: bool = False
                   ) -> pd.Series:

    pdf = calc_logsv_pdf_core(ttm=ttm,
                              sigma0=sigma0,
                              beta=beta,
                              volvol=volvol,
                              log_strikes=log_strikes,
                              is_norm=is_norm,
                              cut=cut,
                              is_analytic=is_analytic)

    if is_norm:
        dx = log_strikes[1] - log_strikes[0]
        pdf = dx * pdf

    return pdf


def calc_logsv_pdf_core(ttm: float,
                        sigma0: float,
                        beta: float,
                        volvol: float,
                        log_strikes: Optional[np.ndarray] = None,
                        is_norm: bool = False,
                        cut: float = 6.0,
                        is_analytic: bool = False
                        ) -> pd.Series:

    """
    compute pdf
    """
    vol_t = sigma0 * np.sqrt(ttm)

    if log_strikes is None:
        log_strikes = np.linspace(-cut * vol_t, cut * vol_t, 100)

    sigma_0, dsigma, d2sigma = calc_logsv_ivols_partials(log_strikes=log_strikes,
                                                         sigma0=sigma0,
                                                         beta=beta,
                                                         volvol=volvol,
                                                         eps=0.001,
                                                         mult=np.sqrt(ttm),
                                                         is_analytic=is_analytic)
    f1 = log_strikes / sigma_0 - 0.5 * sigma_0
    f2 = log_strikes / sigma_0 + 0.5 * sigma_0
    df1 = (1.0-dsigma*f2)/sigma_0
    df2 = (1.0 - dsigma * f1) / sigma_0
    pdf = norm.pdf(f2) * (sigma_0*df1*df2 + d2sigma)
    return pd.Series(pdf, index=log_strikes)


def infer_strikes_from_deltas(deltas: np.ndarray,
                              forward: float,
                              ttm: float,
                              sigma0: float,
                              beta: float,
                              volvol: float
                              ) -> pd.Series:
    """
    givem
    """
    st = np.sqrt(ttm)

    def func(strike: float, given_delta: float) -> float:
        log_strike = np.log(strike / forward)
        vol_st = st * calc_logsv_ivols(log_strikes=log_strike, sigma0=sigma0, beta=beta, volvol=volvol)
        if given_delta >= 0.0:
            target = norm.ppf(given_delta)
        else:
            target = norm.ppf(1.0+given_delta)
        f = -log_strike / vol_st + 0.5 * vol_st - target
        return f

    imp_deltas = {}
    for idx, given_delta in enumerate(deltas):
        try:
            strike = brenth(f=func, a=0.0001 * forward, b=200.0 * forward, args=(given_delta))  # , x0=forward
        except:
            print(f"can't find strike for delta={given_delta}, ttm={ttm}, forward={forward}")
            strike = forward
        imp_deltas[given_delta] = strike

    imp_deltas = pd.DataFrame.from_dict(imp_deltas, orient='index')
    return imp_deltas.iloc[:, 0]


def get_vols_delta_space(forward: float,
                         ttm: float,
                         sigma0: float,
                         beta: float,
                         volvol: float,
                         deltas: Optional[np.ndarray] = None,
                         is_remap_to_str_delta: bool = True
                         ) -> pd.Series:
    if deltas is None:
        deltas = np.linspace(0.01, 0.99, 100)

    imp_strikes = infer_strikes_from_deltas(deltas=deltas,
                                            forward=forward,
                                            ttm=ttm,
                                            sigma0=sigma0,
                                            beta=beta,
                                            volvol=volvol)
    log_strikes = np.log(imp_strikes.to_numpy() / forward)
    vols = calc_logsv_ivols(log_strikes=log_strikes, sigma0=sigma0, beta=beta, volvol=volvol)
    vols = pd.Series(vols, index=deltas).sort_index()
    if is_remap_to_str_delta:
        put_cond = vols.index > 0.5
        put_vols = vols[put_cond]
        put_vols.index = [x-1.0 for x in put_vols.index]
        put_vols=put_vols.sort_index(ascending=False)
        call_vols = vols[put_cond == False].sort_index(ascending=False)
        put_vols.index = [f"{x:0.2f}" for x in put_vols.index]
        call_vols.index = [f"{x:0.2f}" for x in call_vols.index]
        vols = pd.concat([put_vols, call_vols], axis=0)
    return vols


def get_pdf_delta_space(forward: float,
                        ttm: float,
                        sigma0: float,
                        beta: float,
                        volvol: float,
                        deltas: Optional[np.ndarray] = None,
                        is_remap_to_straddle_delta: bool = True,
                        is_analytic: bool = True,
                        ) -> pd.Series:
    if deltas is None:
        deltas = np.linspace(0.01, 0.99, 100)

    imp_strikes = infer_strikes_from_deltas(deltas=deltas,
                                            forward=forward,
                                            ttm=ttm,
                                            sigma0=sigma0,
                                            beta=beta,
                                            volvol=volvol)
    log_strikes = np.log(imp_strikes.to_numpy() / forward)
    pdfs = calc_logsv_pdf_core(ttm=ttm,
                               sigma0=sigma0,
                               beta=beta,
                               volvol=volvol,
                               log_strikes=log_strikes,
                               is_norm=False,
                               is_analytic=is_analytic)
    if is_remap_to_straddle_delta:
        vol_index = -2.0 * deltas + 1.0
    else:
        vol_index = deltas

    return pd.Series(pdfs, index=vol_index).sort_index()


def generate_grid_option_prices_from_slice(vols: pd.Series,
                                           given_log_strikes: np.ndarray,
                                           log_strike_grid: np.ndarray,
                                           p0_ref: float,
                                           ttm: float,
                                           vol_addon: Optional[float] = None
                                           ) -> Tuple[pd.Series, pd.Series]:
    """
    fit slice data and produce grid of option prices
    """
    fit_params = fit_logsv_ivols(log_strikes=given_log_strikes,
                                 mid_vols=vols.to_numpy(),
                                 ttm=ttm,
                                 is_vega_weights=True)

    if vol_addon is not None:
        fit_params[ATM_VOL] += vol_addon

    vols = calc_logsv_ivols(log_strikes=log_strike_grid, **fit_params)
    extended_strikes = p0_ref*np.exp(log_strike_grid)
    calls = bsm.compute_bsm_vanilla_slice_prices(ttm=ttm,
                                                 forward=p0_ref,
                                                 strikes=extended_strikes,
                                                 vols=vols,
                                                 optiontypes=np.full_like(extended_strikes, 'C', dtype=str))

    puts = bsm.compute_bsm_vanilla_slice_prices(ttm=ttm,
                                                forward=p0_ref,
                                                strikes=extended_strikes,
                                                vols=vols,
                                                optiontypes=np.full_like(extended_strikes, 'P', dtype=str))
    calls = pd.Series(calls, index=extended_strikes)
    puts = pd.Series(puts, index=extended_strikes)

    return puts, calls


class UnitTests(Enum):
    PDF = 1
    DELTAS = 2
    DELTA_CHECK = 3
    VOLS_IN_DELTA = 4
    PDFS_IN_DELTA = 5
    FIT = 6
    GENERATE_PRICES = 7


def run_unit_test(unit_test: UnitTests):

    # to get options data
    from option_chain_analytics.data.chain_ts import OptionsDataDFs
    from option_chain_analytics.data.chain_loader_from_ts import create_chain_from_from_options_dfs
    from option_chain_analytics.public_apis.deribit import load_deribit_contract_ts_data

    if unit_test == UnitTests.PDF:
        log_strikes = np.linspace(-5, 5, 1000)
        ttms = np.linspace(0.1, 1.0, 10)

        pdfs = []
        for ttm in ttms:
            pdf = calc_logsv_pdf(ttm=ttm,
                                 sigma0=1.0,
                                 beta=1.0,
                                 volvol=1.0,
                                 log_strikes=log_strikes,
                                 is_norm=False)
            pdfs.append(pdf.rename(f"ttm={ttm:0.2f}"))
        pdfs = pd.concat(pdfs, axis=1)

        with sns.axes_style('darkgrid'):
            fig, ax = plt.subplots(1, 1, figsize=(10, 10), tight_layout=True)
        sns.lineplot(data=pdfs, dashes=False, ax=ax)

    elif unit_test == UnitTests.DELTAS:
        deltas = np.linspace(0.05, 0.95, 9)
        imp_strikes = infer_strikes_from_deltas(deltas=deltas,
                                                forward=1.0,
                                                ttm=1.0,
                                                sigma0=1.0,
                                                beta=-0.5,
                                                volvol=0.5)

        print(imp_strikes)

    elif unit_test == UnitTests.DELTA_CHECK:
        log_strikes = np.linspace(-3.0, 3.0, 100)
        vols = calc_logsv_ivols(log_strikes=log_strikes, sigma0=1.0, beta=0.5, volvol=0.5)
        bsm_deltas = bsm.compute_bsm_vanilla_slice_deltas(ttm=1.0,
                                                          forward=1.0,
                                                          strikes=np.exp(log_strikes),
                                                          vols=vols,
                                                          optiontypes=np.full_like(log_strikes, 'C', dtype=str))
        print(bsm_deltas)
        bsm_deltas = pd.Series(bsm_deltas, index=log_strikes)
        sns.lineplot(data=bsm_deltas)

    elif unit_test == UnitTests.VOLS_IN_DELTA:
        deltas = np.linspace(0.05, 0.95, 100)
        ttms = np.linspace(0.01, 1.0, 10)
        imp_vols = []
        for ttm in ttms:
            imp_vol = get_vols_delta_space(deltas=deltas,
                                           forward=1.0,
                                           ttm=ttm,
                                           sigma0=1.0,
                                           beta=0.0,
                                           volvol=0.5)
            imp_vols.append(imp_vol.rename(f"ttm={ttm: 0.2f}"))

        imp_vols = pd.concat(imp_vols, axis=1)
        sns.lineplot(data=imp_vols)

    elif unit_test == UnitTests.PDFS_IN_DELTA:
        deltas = np.linspace(0.05, 0.95, 100)
        ttms = np.linspace(0.01, 1.0, 10)
        pdfs = []
        for ttm in ttms:
            pdf = get_pdf_delta_space(deltas=deltas,
                                      forward=1.0,
                                      ttm=ttm,
                                      sigma0=1.0,
                                      beta=0.0,
                                      volvol=0.5,
                                      is_analytic=False)
            pdfs.append(pdf.rename(f"ttm={ttm: 0.2f}"))

        imp_vols = pd.concat(pdfs, axis=1)
        sns.lineplot(data=imp_vols)

    elif unit_test == UnitTests.FIT:

        options_data_dfs = OptionsDataDFs(**load_deribit_contract_ts_data(ticker='ETH'))
        time_index = options_data_dfs.get_timeindex()
        print(f"time_index={time_index}")
        value_time = pd.Timestamp('2023-11-01 15:31:57.487172+00:00')
        chain = create_chain_from_from_options_dfs(options_data_dfs=options_data_dfs, value_time=value_time)
        slice_id = '29Dec2023'
        this = chain.get_expiry_slice(slice_id=slice_id)
        vols, strikes = this.get_bid_mark_ask_vols()
        mid_vols = vols['ask_iv'].to_numpy()
        log_strikes = np.log(strikes.to_numpy()/this.forward)

        fit_params = fit_logsv_ivols(log_strikes=log_strikes, mid_vols=mid_vols, ttm=this.get_ttm(), is_vega_weights=True)
        print(fit_params)

    elif unit_test == UnitTests.GENERATE_PRICES:

        options_data_dfs = OptionsDataDFs(**load_deribit_contract_ts_data(ticker='ETH'))
        time_index = options_data_dfs.get_timeindex()
        print(f"time_index={time_index}")
        value_time = pd.Timestamp('2023-11-01 15:31:57.487172+00:00')
        chain = create_chain_from_from_options_dfs(options_data_dfs=options_data_dfs, value_time=value_time)
        e_slice = chain.get_expiry_slice(slice_id='10Nov2023')
        p0_ref = 1800.0
        vols, given_log_strikes = e_slice.get_vols_with_logstrikes()
        vol = 0.4
        ttm = e_slice.get_ttm()
        log_strike_grid = np.linspace(-5.0*vol*np.sqrt(ttm), 0.5*vol*np.sqrt(ttm))

        generate_grid_option_prices_from_slice(vols=vols, given_log_strikes=given_log_strikes,
                                               log_strike_grid=log_strike_grid,
                                               p0_ref=p0_ref,
                                               ttm=ttm)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.FIT

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
create chain data object with options using time series options data in OptionsDataDFs
"""
# package
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from typing import Tuple, Dict, Optional, List
from enum import Enum
import qis
from qis import timer, TimePeriod

# analytics
from option_chain_analytics.option_chain import SliceColumn, UnderlyingColumn, ExpirySlice, SlicesChain
from option_chain_analytics.data.chain_ts import OptionsDataDFs


# @timer
def create_chain_from_from_options_dfs(options_data_dfs: OptionsDataDFs,
                                       value_time: pd.Timestamp,
                                       ) -> Optional[SlicesChain]:

    options_df = options_data_dfs.get_time_slice(timestamp=value_time)

    if not options_df.empty:
        # options_df[SliceColumn.CONTRACT.value] = options_df.index.to_list()
        mat_slice = options_df.groupby(SliceColumn.MATURITY_ID.value)
        expiry_slices, undelying_datas = {}, {}
        for mat, df in mat_slice:
            if np.any(df[SliceColumn.OPEN_INTEREST].isna() == False):
                # forward is contract weighted avs
                forward = qis.np_nonan_weighted_avg(a=df[SliceColumn.UNDERLYING_PRICE],
                                                    weights=df[SliceColumn.OPEN_INTEREST])
            else:
                # forward is contract weighted avs
                forward = np.nanmean(df[SliceColumn.UNDERLYING_PRICE])
            if not np.isnan(forward):
                undelying_data = {UnderlyingColumn.EXPIRY_ID: str(mat),
                                  UnderlyingColumn.VALUE_TIME: value_time,
                                  UnderlyingColumn.EXPIRY: df[SliceColumn.EXPIRY].iloc[0],
                                  UnderlyingColumn.SPOT_PRICE: forward,  # to do
                                  UnderlyingColumn.UNDERLYING_INDEX: str(mat),  # to do
                                  UnderlyingColumn.FORWARD_PRICE: forward,  # to do
                                  UnderlyingColumn.IR_RATE: 0.0,
                                  UnderlyingColumn.TTM: df[SliceColumn.TTM].iloc[0]}
                undelying_data = pd.Series(undelying_data)
                expiry_slices[str(mat)] = ExpirySlice(options_df=df, undelying_data=undelying_data)
                undelying_datas[str(mat)] = undelying_data
        undelying_df = pd.DataFrame.from_dict(undelying_datas, orient='index')
        chain = SlicesChain(options_df=options_df,#.set_index(SliceColumn.CONTRACT),
                            undelying_df=undelying_df,
                            expiry_slices=expiry_slices,
                            value_time=value_time)
    else:
        chain = None
    return chain


@timer
def create_chain_timeseries(options_data_dfs: OptionsDataDFs,
                            dates_schedule: pd.DatetimeIndex = None,
                            time_period: TimePeriod = None,
                            freq: str = 'W-FRI',
                            hour_offset: int = 8
                            ) -> Dict[pd.Timestamp, SlicesChain]:

    if dates_schedule is None:
        if time_period is None:
            raise ValueError(f"time_period={time_period} must be non Nons")
        dates_schedule = qis.generate_dates_schedule(time_period=time_period,
                                                     freq=freq,
                                                     hour_offset=hour_offset)

    chain_data = {}
    for timestamp in dates_schedule:
        chain = create_chain_from_from_options_dfs(options_data_dfs=options_data_dfs, value_time=timestamp)
        if chain is not None:
            chain_data[timestamp] = chain
    return chain_data


@timer
def generate_atm_vols(options_data_dfs: OptionsDataDFs,
                      time_period: TimePeriod = None,
                      freq: str = 'D',
                      hour_offset: int = 8,
                      days_before_roll: int = 7
                      ) -> pd.Series:

    if time_period is None:
        time_period = options_data_dfs.get_start_end_date()

    chain_data = create_chain_timeseries(options_data_dfs=options_data_dfs,
                                         time_period=time_period,
                                         freq=freq,
                                         hour_offset=hour_offset)
    vols = {}
    for date, chain in chain_data.items():
        next_date = date + pd.DateOffset(days=days_before_roll)
        slice_id = chain.get_next_slice_after_date(mat_date=next_date)
        vols[date] = chain.get_atm_vol(slice_id=slice_id)
    vols = pd.Series(vols)
    return vols


@timer
def generate_vol_delta_ts(options_data_dfs: OptionsDataDFs,
                          days_map: Dict[str, int] = None,
                          deltas: List[float] = (-0.10, -0.25, 0.50, 0.25, 0.10),
                          freq: str = 'D',
                          hour_offset: Optional[int] = 8,
                          time_period: TimePeriod = None
                          ) -> Tuple[pd.DataFrame, ...]:

    if time_period is None:
        time_period = options_data_dfs.get_start_end_date()

    chain_data = create_chain_timeseries(options_data_dfs=options_data_dfs,
                                         time_period=time_period,
                                         freq=freq,
                                         hour_offset=hour_offset)
    if days_map is None:
        days_map = {'1d': 1, '2d': 2, '1w': 7, '2w': 10, '1m': 28, '2m': 56, '3m': 84, '2Q': 168}

    is_add_tenor_to_delta_name = False if len(days_map.keys()) == 1 else True
    vols, strikes, options, underlying_prices = {}, {}, {}, {}
    for date, chain in chain_data.items():
        delta_vol_matrix = chain.generate_delta_vol_matrix(value_time=date, days_map=days_map, deltas=deltas)
        if delta_vol_matrix is not None:
            vols_, strikes_, options_, index_prices_ = delta_vol_matrix.get_melted_matrix(is_add_tenor_to_delta_name=is_add_tenor_to_delta_name)
            vols[date] = vols_['vols'].rename(date)
            strikes[date] = strikes_['strikes'].rename(date)
            options[date] = options_['options'].rename(date)
            underlying_prices[date] = index_prices_['underlying_prices_matrix'].rename(date)
    vols = pd.DataFrame.from_dict(vols, orient='index')
    strikes = pd.DataFrame.from_dict(strikes, orient='index')
    options = pd.DataFrame.from_dict(options, orient='index')
    underlying_prices = pd.DataFrame.from_dict(underlying_prices, orient='index')
    return vols, strikes, options, underlying_prices


class UnitTests(Enum):
    CREATE_CHAIN_AT_TS = 1
    CREATE_TS_CHAIN_DATA = 2
    CREATE_WEEKLY_ROLLS = 3
    GENERATE_VOL_DELTA_TS = 4


def run_unit_test(unit_test: UnitTests):

    from option_chain_analytics.ts_loaders import ts_data_loader_wrapper, DataSource

    ticker = 'ETH'
    options_data_dfs = OptionsDataDFs(**ts_data_loader_wrapper(ticker=ticker, data_source=DataSource.TARDIS_LOCAL))

    if unit_test == UnitTests.CREATE_CHAIN_AT_TS:
        date = pd.Timestamp('2022-09-19 08:00:00+00:00')
        chain = create_chain_from_from_options_dfs(options_data_dfs=options_data_dfs, value_time=date)
        chain.save_joint_slices(file_name=f"{ticker}_{date.strftime('%Y%m%d_%H_%M_%S')}")

        next_date = date+pd.DateOffset(days=30)
        slice_id = chain.get_next_slice_after_date(mat_date=next_date)
        print(f"{date}: {slice_id}")
        print(f"{chain.get_atm_put_id(slice_id=slice_id)}, {chain.get_atm_call_id(slice_id=slice_id)}")
        print(f"{chain.get_put_delta_option_id(slice_id=slice_id, delta=-0.25)}")
        print(f"{chain.get_call_delta_option_id(slice_id=slice_id, delta=0.25)}")

        # days_map = {'1d': 1, '2d': 2, '1w': 7, '2w': 14, '1m': 30, '2m': 60, '3m': 90, '2Q': 120,  '3Q': 180}
        # days_map = {'1d': 1, '2d': 2, '1w': 7, '2w': 14, '1m': 30, '2m': 60, '3m': 90, '2Q': 120}
        days_map = {'2d': 2, '1w': 7, '2w': 14, '1m': 30, '2m': 60, '3m': 90}
        delta_vol_matrix = chain.generate_delta_vol_matrix(value_time=date, days_map=days_map)
        delta_vol_matrix.print()

        vols, strikes, options, index_prices = delta_vol_matrix.get_melted_matrix()
        print(vols)
        print(strikes)
        delta_vol_matrix.plot_vol_in_strike()

    elif unit_test == UnitTests.CREATE_TS_CHAIN_DATA:
        chain_data = create_chain_timeseries(options_data_dfs=options_data_dfs,
                                             time_period=options_data_dfs.get_start_end_date(),
                                             freq='W-FRI',
                                             hour_offset=8)
        for key, chain in chain_data.items():
            print(f"{key}, {chain.expiry_slices.keys()}")

    elif unit_test == UnitTests.CREATE_WEEKLY_ROLLS:
        weekly_fridays_rolls = qis.generate_dates_schedule(TimePeriod(pd.Timestamp('2022-05-06 00:00:00+00:00'),
                                                                      qis.get_current_time_with_tz(tz='UTC', days_offset=7)),
                                                           freq='W-FRI',
                                                           hour_offset=8)
        weekly_fridays_rolls = weekly_fridays_rolls[:-2]
        print(weekly_fridays_rolls)

        chain_data = create_chain_timeseries(options_data_dfs=options_data_dfs,
                                             dates_schedule=weekly_fridays_rolls[:-1])

        for date, next_date in zip(weekly_fridays_rolls[:-1], weekly_fridays_rolls[1:]):
            print(f"{date}: {chain_data[date].get_next_slice_after_date(mat_date=next_date)}, "
                  f"{chain_data[date].get_atm_put_id(mat_date=next_date)}")

    elif unit_test == UnitTests.GENERATE_VOL_DELTA_TS:
        time_period = TimePeriod('2023-01-01 00:00:00+00:00', '2023-07-14 00:00:00+00:00', tz='UTC')
        vols, strikes, options, index_prices = generate_vol_delta_ts(options_data_dfs=options_data_dfs,
                                                                     days_map={'1w': 7, '1m': 30},
                                                                     deltas=[-0.10, -0.25, 0.50, 0.25, 0.10],
                                                                     freq='B',
                                                                     hour_offset=8,
                                                                     time_period=time_period)
        print(vols)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.GENERATE_VOL_DELTA_TS

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)
